# coding: utf-8
"""REST-style HTTP client wrapper over requests."""

__author__ = "Bartek Kryza"
__copyright__ = "Copyright (C) 2023 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"

import json
import logging
from typing import Any, Dict, Optional

import requests
from requests.structures import CaseInsensitiveDict

from .errors import OnedataRESTError

_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())

_HEADERS_TO_LOG = {"range"}


class HttpClient:
    """REST-style wrapper over requests library."""

    timeout: int = 8
    session: requests.Session

    def __init__(self, *, verify_ssl: bool = True) -> None:
        """Construct OnedataFileClient instance."""
        self.session = requests.Session()
        self.session.verify = verify_ssl

    def get_session(self) -> requests.Session:
        """Return requests session instance."""
        return self.session

    def get(
        self,
        url: str,
        data: Any = None,
        headers: Optional[Dict[str, str]] = None,
        *,
        stream: bool = False,
    ) -> requests.Response:
        """Perform a GET request."""
        return self._send_request("GET", url, data, headers, stream=stream)

    def put(
        self, url: str, data: Any = None, headers: Optional[Dict[str, str]] = None
    ) -> requests.Response:
        """Perform a PUT request."""
        return self._send_request("PUT", url, data, headers)

    def post(
        self, url: str, data: Any = None, headers: Optional[Dict[str, str]] = None
    ) -> requests.Response:
        """Perform a POST request."""
        return self._send_request("POST", url, data, headers)

    def delete(
        self, url: str, data: Any = None, headers: Optional[Dict[str, str]] = None
    ) -> requests.Response:
        """Perform a DELETE request."""
        return self._send_request("DELETE", url, data, headers)

    def head(
        self, url: str, data: Any = None, headers: Optional[Dict[str, str]] = None
    ) -> requests.Response:
        """Perform a HEAD request."""
        return self._send_request("HEAD", url, data, headers)

    def _send_request(
        self,
        method: str,
        url: str,
        data: Any = None,
        headers: Optional[Dict[str, str]] = None,
        *,
        stream: bool = False,
    ) -> requests.Response:
        """Perform an HTTP request."""
        if isinstance(data, dict):
            body = json.dumps(data)
        else:
            body = data

        case_insensitive_headers = CaseInsensitiveDict(headers)
        if "content-type" not in case_insensitive_headers:
            case_insensitive_headers["content-type"] = "application/json"

        req = requests.Request(method, url, data=body, headers=case_insensitive_headers)
        prepared = self.session.prepare_request(req)
        response = self.session.send(prepared, stream=stream, timeout=self.timeout)

        if not response.ok:
            ex = OnedataRESTError.from_response(response)
            if response.status_code == 403:
                _logger.debug(
                    "AUTHORIZATION FAILED %s %s\nheaders: %s\nbody: %s\ndetails: %s",
                    method,
                    url,
                    {
                        k: v
                        for k, v in case_insensitive_headers.items()
                        if k.lower() in _HEADERS_TO_LOG
                    },
                    data if isinstance(data, dict) else len(data),
                    ex,
                )
            else:
                _logger.debug("REQUEST FAILED %s %s\ndetails: %s", method, url, ex)

            raise ex

        _logger.debug("REQUEST OK %d %s %s", response.status_code, method, url)
        return response



#!/usr/bin/env python3

"""Runs integration tests."""

import argparse
import os
import platform
import subprocess
import sys
import time

script_dir = os.path.dirname(os.path.realpath(__file__))
docker_dir = os.path.join("bamboos", "docker")
sys.path.insert(0, docker_dir)
from environment import docker, dockers_config

parser = argparse.ArgumentParser(
    formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    description="Run Common Tests.",
)

parser.add_argument(
    "--onenv-config",
    action="store",
    help="run tests in one-env environment",
    dest="onenv_config",
)

parser.add_argument(
    "--no-clean",
    action="store_true",
    default=False,
    help="do not clean the one-env deployment after test",
    dest="no_clean",
)

parser.add_argument(
    "--image",
    "-i",
    action="store",
    default="onedata/pybuilder:v2",
    help="docker image to use as a test master",
    dest="image",
)

parser.add_argument(
    "--suite", action="append", default=[], help="name of the test suite", dest="suites"
)

parser.add_argument(
    "--no-shed-privileges",
    action="store_true",
    default=False,
    help="Run tests as root in container",
    dest="no_shed_privileges",
)


[args, _pass_args] = parser.parse_known_args()
dockers_config.ensure_image(args, "image", "builder")


script_dir = os.path.dirname(os.path.realpath(__file__))

base_test_dir = os.path.realpath("tests")

test_dirs = map(lambda suite: os.path.join(base_test_dir, suite), args.suites)
if not test_dirs:
    test_dirs = [base_test_dir]

envs = {
    "BASE_TEST_DIR": base_test_dir,
    "PYTHONWARNINGS": "ignore:Unverified HTTPS request",
}

# Setup oneenv environment
if args.onenv_config is not None:
    if not os.path.exists(args.onenv_config):
        print(f"Error: No such one-env file {args.onenv_config}")
        sys.exit(1)

    branch = os.popen("git rev-parse --abbrev-ref HEAD").read().strip()
    if branch.startswith("release/"):
        release_tag = branch.split("/", maxsplit=1)[-1]

        with open(args.onenv_config, "r+") as f:
            content = f.read()
            for repo in ("onezone-dev", "oneprovider-dev"):
                develop_image = f"docker.onedata.org/{repo}:develop"
                release_image = f"docker.onedata.org/{repo}:{release_tag}"
                content = content.replace(develop_image, release_image)
            f.seek(0)
            f.truncate()
            f.write(content)

    try:
        up_output = subprocess.check_output(
            ["./one-env/onenv", "up", args.onenv_config]
        )
    except subprocess.CalledProcessError as e:
        print(f"Failed to start onenv up due to: {e.output}")
        sys.exit(1)

    environment_ready = False
    retries = 3
    while (not environment_ready) and retries > 0:
        try:
            subprocess.check_call(["./one-env/onenv", "wait"])
            environment_ready = True
        except subprocess.CalledProcessError as e:
            retries -= 1
            time.sleep(5)
            print("Waiting for one-env environment setup...")

    if retries == 0:
        print(f"Failed to start K8S environment from {args.onenv_config}")
        sys.exit(1)

    print("One-env environment ready")

    # Get Onezone IP
    get_onezoneip_cli = "kubectl get pod dev-onezone-0 --template {{.status.podIP}}"
    onezone_ip = subprocess.check_output(get_onezoneip_cli.split(" ")).strip()
    envs["DEV_ONEZONE_0"] = onezone_ip.decode("utf-8")

    # Get Oneprovider IP
    get_oneproviderip_cli = (
        "kubectl get pod dev-oneprovider-krakow-0 --template {{.status.podIP}}"
    )
    oneprovider_ip = subprocess.check_output(get_oneproviderip_cli.split(" ")).strip()
    envs["DEV_ONEPROVIDER_KRAKOW_0"] = oneprovider_ip.decode("utf-8")

    get_oneprovider2ip_cli = (
        "kubectl get pod dev-oneprovider-paris-0 --template {{.status.podIP}}"
    )
    oneprovider_2_ip = subprocess.check_output(
        get_oneprovider2ip_cli.split(" ")
    ).strip()
    envs["DEV_ONEPROVIDER_PARIS_0"] = oneprovider_2_ip.decode("utf-8")

    print(f"Environment passed to pytest container: {str(envs)}")


command = """
import os, subprocess, sys, stat, shutil

if {shed_privileges}:
    os.environ['HOME'] = '/tmp'
    docker_gid = os.stat('/var/run/docker.sock').st_gid
    os.chmod('/etc/resolv.conf', 0o666)
    os.setgroups([docker_gid])
    os.setregid({gid}, {gid})
    os.setreuid({uid}, {uid})

tests = 'tests'
if '{suites}':
    tests = '{suites}'

command = ['python3'] + ['-m'] + ['tox'] +  ['-c'] + ['tox.ini'] + ['-e'] + [tests]
ret = subprocess.call(command)
sys.exit(ret)
"""
command = command.format(
    uid=os.geteuid(),
    gid=os.getegid(),
    shed_privileges=(platform.system() == "Linux") and not args.no_shed_privileges,
    suites=",".join(args.suites),
)

ret = docker.run(
    tty=True,
    rm=True,
    interactive=True,
    workdir=script_dir,
    volumes=[(script_dir, script_dir, "rw")],
    reflect=[("/var/run/docker.sock", "rw")],
    image=args.image,
    envs=envs,
    run_params=["--privileged"] if args.no_shed_privileges else [],
    command=["python3", "-c", command],
)

if not args.no_clean:
    try:
        up_output = subprocess.check_output(["./one-env/onenv", "clean"])
    except subprocess.CalledProcessError as e:
        print(f"Failed to clean onenv due to: {e.output}")
        sys.exit(1)

sys.exit(ret)



# coding: utf-8
"""Common test functions and fixtures."""

__author__ = "Bartek Kryza"
__copyright__ = "Copyright (C) 2023 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"

import random
import string
from typing import Optional


def random_int(lower_bound: int = 1, upper_bound: int = 100) -> int:
    """Generate random int in a given range."""
    return random.randint(lower_bound, upper_bound)


def random_str(
    size: Optional[int] = None,
    *,
    characters: str = string.ascii_uppercase + string.digits
) -> str:
    """Generate random string of specified size."""
    size = size or random_int()
    return "".join(random.choice(characters) for _ in range(size))


def random_path(size: Optional[int] = None) -> str:
    """Generate random file system path."""
    size = size or random_int(3, 10)
    return "/".join(random_str(5) for _ in range(size))


def random_bytes(size: Optional[int] = None) -> bytes:
    """Generate random sequence of bytes."""
    size = size or random_int()
    return random_str(size).encode("utf-8")



# coding: utf-8
"""Definitions of environment constants used in tests."""

__author__ = "Bartek Kryza"
__copyright__ = "Copyright (C) 2023 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"


from typing import Final

ZONE_DOMAIN: Final[str] = "dev-onezone-0.default.svc.cluster.local"

PROVIDER_KRK_NAME: Final[str] = "dev-oneprovider-krakow"
PROVIDER_KRK_DOMAIN: Final[str] = "dev-oneprovider-krakow.default.svc.cluster.local"
PROVIDER_PAR_NAME: Final[str] = "dev-oneprovider-paris"
PROVIDER_PAR_DOMAIN: Final[str] = "dev-oneprovider-paris.default.svc.cluster.local"

SPACE_KRK_PAR_NAME: Final[str] = "space_krk_par"
SPACE_PAR_NAME: Final[str] = "space_par"
SPACE_NO_SUPPORT_NAME: Final[str] = "space_nosupport"

ADMIN_USERNAME: Final[str] = "admin"
ADMIN_PASSWORD: Final[str] = "password"
SPACE_MEMBER_USERNAME: Final[str] = "space_member"
SPACE_MEMBER_PASSWORD: Final[str] = "password"



# coding: utf-8
"""Pytest test setup."""

__author__ = "Bartek Kryza"
__copyright__ = "Copyright (C) 2023 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"

import logging
import os
import time
from typing import Final

import requests
from urllib3.util import connection

import pytest

from . import (
    ADMIN_PASSWORD,
    ADMIN_USERNAME,
    PROVIDER_KRK_DOMAIN,
    PROVIDER_PAR_DOMAIN,
    SPACE_MEMBER_PASSWORD,
    SPACE_MEMBER_USERNAME,
    ZONE_DOMAIN,
)


def trace_requests_messages() -> None:
    """Enable logging HTTP requests."""
    # pylint: disable=C0415
    import http.client as http_client

    http_client.HTTPConnection.debuglevel = 1

    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


# Uncomment to enable HTTP request trace log
# trace_requests_messages()


_original_create_connection = connection.create_connection


def patched_create_connection(address, *args, **kwargs):
    """Resolve Kubernetes domain names to IP's from environment variables."""
    host, port = address
    hostname = host

    if host == ZONE_DOMAIN:
        hostname = os.getenv("DEV_ONEZONE_0")
    elif host == PROVIDER_KRK_DOMAIN:
        hostname = os.getenv("DEV_ONEPROVIDER_KRAKOW_0")
    elif host == PROVIDER_PAR_DOMAIN:
        hostname = os.getenv("DEV_ONEPROVIDER_PARIS_0")

    return _original_create_connection((hostname, port), *args, **kwargs)


connection.create_connection = patched_create_connection


FIXTURE_SCOPE: Final[str] = "session"


@pytest.fixture(scope="module", autouse=True)
def fixture_wait_for_support_sync():
    """Wait until providers are fully synchronized after setup."""
    time.sleep(10)


@pytest.fixture(scope=FIXTURE_SCOPE, name="onezone_ip")
def fixture_onezone_ip():
    """Get Onezone IP from environment variable."""
    return os.getenv("DEV_ONEZONE_0")


@pytest.fixture(scope=FIXTURE_SCOPE, name="provider_krk_ip")
def fixture_provider_krk_ip():
    """Get Oneprovider 'krakow' IP from environment variable."""
    return os.getenv("DEV_ONEPROVIDER_KRAKOW_0")


@pytest.fixture(scope=FIXTURE_SCOPE, name="provider_par_ip")
def fixture_provider_par_ip():
    """Get Oneprovider 'paris' IP from environment variable."""
    return os.getenv("DEV_ONEPROVIDER_PARIS_0")


@pytest.fixture(scope=FIXTURE_SCOPE, name="onezone_admin_token")
def fixture_onezone_admin_token(onezone_ip):
    """Generate a new admin token."""
    return _create_temp_token(onezone_ip, ADMIN_USERNAME, ADMIN_PASSWORD)


@pytest.fixture(scope=FIXTURE_SCOPE, name="onezone_readonly_token")
def fixture_onezone_readonly_token(onezone_ip):
    """Generate new readonly only admin token."""
    return _create_temp_token(onezone_ip, ADMIN_USERNAME, ADMIN_PASSWORD, readonly=True)


@pytest.fixture(scope=FIXTURE_SCOPE, name="onezone_space_member_token")
def fixture_space_member_token(onezone_ip):
    """Generate a new space member token."""
    return _create_temp_token(onezone_ip, SPACE_MEMBER_USERNAME, SPACE_MEMBER_PASSWORD)


def _create_temp_token(
    onezone_ip: str, username: str, password: str, *, readonly: bool = False
) -> str:
    url = f"https://{onezone_ip}/api/v3/onezone/user/tokens/temporary"
    headers = {"content-type": "application/json"}
    auth = requests.auth.HTTPBasicAuth(username, password)

    caveats = [{"type": "time", "validUntil": int(time.time()) + 2590000}]
    if readonly:
        caveats.append({"type": "data.readonly"})

    result = requests.post(
        url,
        json={
            "type": {"accessToken": {}},
            "caveats": caveats,
        },
        headers=headers,
        auth=auth,
        verify=False,
    )
    return result.json()["token"]



# coding: utf-8
"""Onedata REST file API client."""

from __future__ import annotations

import json
import sys
import typing
from functools import partial, wraps
from typing import Any, Callable, Dict, Iterator, List, Optional

import requests.exceptions

from .errors import NoAvailableProviderForSpaceError, ReadonlyTokenError
from .file_attributes import (
    FileAttrKey,
    FileAttrsJson,
    FileType,
    build_http_get_file_attr_params,
    build_http_list_children_attrs_params,
    normalize_file_attrs_json,
)
from .http_client import HttpClient
from .onezone_rest_client import (
    AccessTokenScope,
    OnezoneRESTClient,
    SpaceFQN,
    SpaceId,
    SpaceSpecifier,
)
from .provider_selector import (
    ProviderSelector,
    ProviderSpecifier,
    SpaceSupportingProvider,
)

__author__ = "Bartek Kryza"
__copyright__ = "Copyright (C) 2023 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"


if sys.version_info < (3, 11):
    from typing_extensions import TypeAlias, TypedDict
else:
    from typing import TypeAlias, TypedDict


FileId: TypeAlias = str
FilePath: TypeAlias = str
"""
File path relative to space, that is without space specifier prefix.
"""


class ListChildrenResult(TypedDict):
    """Directory listing result.

    Refer to the API specification for more information:
    https://onedata.org/#/home/api/stable/oneprovider?anchor=operation/list_children
    """

    children: List[FileAttrsJson]
    isLast: bool
    nextPageToken: Optional[str]


def _find_available_provider(
    func: Optional[Callable[..., Any]] = None, *, except_readonly: bool = False
) -> Callable[..., Any]:
    if func is None:
        return partial(_find_available_provider, except_readonly=except_readonly)

    @wraps(func)
    def wrapper(
        self: OnedataFileRESTClient,
        space_specifier: SpaceSpecifier,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        if except_readonly:
            access_token_scope = self.get_token_scope()
            if access_token_scope["dataAccessScope"]["readonly"]:
                raise ReadonlyTokenError

        # pylint: disable=W0212
        oz_client = self._oz_client
        provider_selector = self._provider_selector

        space_fqn = oz_client.ensure_space_fqn(space_specifier)

        provider = kwargs.get("provider")
        if provider is not None:
            return func(self, space_fqn, *args, **kwargs)

        for provider in provider_selector.iter_available_space_providers(
            space_fqn,
            oz_rest_client=oz_client,
            except_readonly=except_readonly,
        ):
            try:
                kwargs["provider"] = provider
                return func(self, space_fqn, *args, **kwargs)
            except (
                requests.exceptions.ConnectionError,
                requests.exceptions.ReadTimeout,
            ):
                provider_selector.blacklist(provider)

        raise NoAvailableProviderForSpaceError(space_specifier)

    return wrapper


class OnedataFileRESTClient:
    """Custom REST client for Onedata REST basic file operations API."""

    _oz_client: OnezoneRESTClient
    _provider_selector: ProviderSelector
    _op_client: HttpClient

    def __init__(
        self,
        onezone_host: str,
        token: str,
        preferred_providers: Optional[List[ProviderSpecifier]] = None,
        *,
        verify_ssl: bool = True,
    ):
        """Construct OnedataFileRESTClient instance."""
        self._oz_client = OnezoneRESTClient(
            host=onezone_host, token=token, verify_ssl=verify_ssl
        )
        self._provider_selector = ProviderSelector(
            preferred_providers=preferred_providers
        )

        self._op_client = HttpClient(verify_ssl=verify_ssl)
        self._op_client.get_session().headers.update({"X-Auth-Token": token})

    def __eq__(self, other: object) -> bool:
        """Compare 2 instances of OnedataFileRESTClient."""
        if not isinstance(other, OnedataFileRESTClient):
            return NotImplemented

        return self._oz_client == other._oz_client

    def __hash__(self) -> int:
        """Calculate a hash of a given instance of OnedataFileRESTClient."""
        return hash(self._oz_client)

    def get_token_scope(self) -> AccessTokenScope:
        """Get current token access scope."""
        return self._oz_client.infer_token_scope()

    def list_spaces(self) -> List[SpaceFQN]:
        """List all spaces available for the current token."""
        return self._oz_client.list_spaces()

    def get_space_id(self, space_specifier: SpaceSpecifier) -> SpaceId:
        """Get space id by specifier."""
        return self._oz_client.get_space_id(space_specifier)

    @_find_available_provider
    def get_file_id(
        self,
        space_specifier: SpaceSpecifier,
        *,
        file_path: Optional[FilePath] = None,
        retries: int = 3,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> FileId:
        """Get Onedata file id based on space specifier and file path."""
        provider = self._ensure_provider(provider)

        rest_path = f"/lookup-file-id/{space_specifier}"
        if file_path is not None:
            rest_path += f"/{file_path.lstrip('/')}"
        url = self._build_op_url(provider, rest_path)

        while True:
            try:
                file_id = self._op_client.post(url).json()["fileId"]
                return typing.cast(FileId, file_id)
            except requests.exceptions.ReadTimeout as e:
                if retries <= 0:
                    raise e

                retries -= 1

    @_find_available_provider
    def get_attributes(
        self,
        space_specifier: SpaceSpecifier,
        *,
        attributes: Optional[List[FileAttrKey]] = None,
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> FileAttrsJson:
        """Get file or directory attributes."""
        provider = self._ensure_provider(provider)
        file_id = self._resolve_file_id(
            space_specifier, file_path=file_path, file_id=file_id, provider=provider
        )
        url = self._build_op_url(provider, f"/data/{file_id}")
        qs, body = build_http_get_file_attr_params(provider, attributes)
        if qs:
            url += f"?{qs}"

        result = self._op_client.get(url, data=body).json()
        attrs = normalize_file_attrs_json(provider, attributes, result)

        return typing.cast(FileAttrsJson, attrs)

    @_find_available_provider(except_readonly=True)
    def set_attributes(
        self,
        space_specifier: SpaceSpecifier,
        attributes: Dict[str, str],
        *,
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> None:
        """Set file or directory attributes."""
        provider = self._ensure_provider(provider)
        file_id = self._resolve_file_id(
            space_specifier, file_path=file_path, file_id=file_id, provider=provider
        )
        url = self._build_op_url(provider, f"/data/{file_id}")
        self._op_client.put(url, data=attributes)

    @_find_available_provider
    def list_children(
        self,
        space_specifier: SpaceSpecifier,
        *,
        attributes: Optional[List[FileAttrKey]] = None,
        limit: int = 1000,
        continuation_token: Optional[str] = None,
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> ListChildrenResult:
        """List contents of a directory."""
        provider = self._ensure_provider(provider)
        dir_file_id = self._resolve_file_id(
            space_specifier, file_path=file_path, file_id=file_id, provider=provider
        )
        qs = f"?limit={limit}"
        if continuation_token is not None:
            qs += f"&token={continuation_token}"

        if not attributes:
            attributes = ["name", "type"]

        qs_attrs, body = build_http_list_children_attrs_params(provider, attributes)
        if qs_attrs:
            qs += f"&{qs_attrs}"

        url = self._build_op_url(provider, f"/data/{dir_file_id}/children{qs}")
        result = self._op_client.get(url, data=body).json()
        result["children"] = [
            normalize_file_attrs_json(provider, attributes, file_attrs_json)
            for file_attrs_json in result["children"]
        ]

        return typing.cast(ListChildrenResult, result)

    @_find_available_provider
    def get_file_content(
        self,
        space_specifier: SpaceSpecifier,
        *,
        offset: int = 0,
        size: Optional[int] = None,
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> bytes:
        """Read from a file."""
        provider = self._ensure_provider(provider)
        file_id = self._resolve_file_id(
            space_specifier, file_path=file_path, file_id=file_id, provider=provider
        )
        if size is not None:
            headers = {"Range": f"bytes={offset}-{offset + size - 1}"}
        else:
            headers = {}

        url = self._build_op_url(provider, f"/data/{file_id}/content")
        return self._op_client.get(url, headers=headers).content

    @_find_available_provider
    def iter_file_content(
        self,
        space_specifier: SpaceSpecifier,
        *,
        chunk_size: int = 1048576,  # 1 MB
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> Iterator[bytes]:
        """Iterate file content."""
        provider = self._ensure_provider(provider)
        file_id = self._resolve_file_id(
            space_specifier, file_path=file_path, file_id=file_id, provider=provider
        )
        url = self._build_op_url(provider, f"/data/{file_id}/content")
        return self._op_client.get(url, stream=True).iter_content(chunk_size)

    @_find_available_provider(except_readonly=True)
    def put_file_content(
        self,
        space_specifier: SpaceSpecifier,
        data: bytes,
        *,
        offset: Optional[int] = None,
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> None:
        """Write to a file."""
        provider = self._ensure_provider(provider)
        file_id = self._resolve_file_id(
            space_specifier, file_path=file_path, file_id=file_id, provider=provider
        )
        headers = {"Content-type": "application/octet-stream"}
        qs = f"?offset={offset}" if offset is not None else ""
        url = self._build_op_url(provider, f"/data/{file_id}/content{qs}")
        self._op_client.put(url, data=data, headers=headers)

    @_find_available_provider(except_readonly=True)
    def create_file(
        self,
        space_specifier: SpaceSpecifier,
        file_path: FilePath,
        *,
        file_type: FileType = "REG",
        create_parents: bool = False,
        mode: Optional[int] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> FileId:
        """Create a file at path."""
        provider = self._ensure_provider(provider)
        space_id = self.get_space_id(space_specifier)
        parents = str(create_parents).lower()

        path = f"/data/{space_id}/path/{file_path}"
        path += f"?type={file_type}&create_parents={parents}"

        if mode:
            path += f"&mode={oct(mode).lstrip('0o')}"

        url = self._build_op_url(provider, path)
        result = self._op_client.put(url, b"").json()["fileId"]
        return typing.cast(FileId, result)

    @_find_available_provider(except_readonly=True)
    def remove(
        self,
        space_specifier: SpaceSpecifier,
        *,
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> None:
        """Remove a file or directory."""
        provider = self._ensure_provider(provider)
        file_id = self._resolve_file_id(
            space_specifier, file_path=file_path, file_id=file_id, provider=provider
        )
        url = self._build_op_url(provider, f"/data/{file_id}")
        self._op_client.delete(url)

    @_find_available_provider(except_readonly=True)
    def move(
        self,
        src_space_specifier: SpaceSpecifier,
        src_file_path: FilePath,
        dst_space_specifier: SpaceSpecifier,
        dst_file_path: FilePath,
        *,
        provider: Optional[SpaceSupportingProvider] = None,
    ) -> None:
        """Rename a file or directory."""
        provider = self._ensure_provider(provider)
        headers = {
            "X-CDMI-Specification-Version": "1.1.1",
            "Content-type": "application/cdmi-object",
        }
        url = f"https://{provider.domain}/cdmi/{dst_space_specifier}/{dst_file_path}"

        data = {"move": f"{src_space_specifier}/{src_file_path}"}

        self._op_client.put(url, data=json.dumps(data), headers=headers)

    @staticmethod
    def _ensure_provider(
        provider: Optional[SpaceSupportingProvider],
    ) -> SpaceSupportingProvider:
        if provider is None:
            raise ValueError("Missing required 'provider' argument.")

        return provider

    def _resolve_file_id(
        self,
        space_specifier: SpaceSpecifier,
        *,
        file_path: Optional[FilePath] = None,
        file_id: Optional[FileId] = None,
        provider: SpaceSupportingProvider,
    ) -> FileId:
        if file_id is not None:
            return file_id

        if file_path is not None:
            file_id = self.get_file_id(
                space_specifier, file_path=file_path, provider=provider
            )
            return typing.cast(FileId, file_id)

        return self.get_space_id(space_specifier)

    @staticmethod
    def _build_op_url(provider: SpaceSupportingProvider, rest_path: str) -> str:
        if not rest_path.startswith("/"):
            rest_path = "/" + rest_path

        return f"https://{provider.domain}/api/v3/oneprovider{rest_path}"



# coding: utf-8
"""Onedata REST file API client errors module."""

from __future__ import annotations

from typing import Dict, Optional, Union

import requests

__author__ = "Bartek Kryza"
__copyright__ = "Copyright (C) 2023 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"


class OnedataError(Exception):
    """Base exception class for Onedata errors."""

    msg = "Unexpected error"

    def __str__(self) -> str:
        """Return the error message."""
        return self.msg.format(**self.__dict__)


class ReadonlyTokenError(OnedataError):
    """Exception raised when write operation is called with readonly token."""

    msg = (
        "Forbidden operation - the provided access token is limited "
        "to read-only data access."
    )


class SpaceNotFoundError(OnedataError):
    """Exception raised when space is not found in Onedata system."""

    msg = (
        "Could not find the requested Onedata space '{space_specifier}'; "
        "either it does not exist, or the provided access token is limited "
        "to a different subset of spaces."
    )

    def __init__(self, space_specifier: str):
        """Construct 'SpaceNotFoundError' instance."""
        self.space_specifier = space_specifier

        super().__init__(space_specifier)


class NoAvailableProviderForSpaceError(OnedataError):
    """Exception raised when no available provider is found for a space."""

    msg = (
        "Could not find any available Onedata provider for the space '{space_specifier}'; "
        "either the space is not supported at all, or all supporting providers "
        "are offline, malfunctioning or outdated (version >= 21.02.1 is required)."
    )

    def __init__(self, space_specifier: str):
        """Construct 'NoAvailableProviderForSpaceError' instance."""
        self.space_specifier = space_specifier

        super().__init__(space_specifier)


class OnedataRESTError(OnedataError):
    """Custom Onedata REST exception class."""

    def __init__(
        self,
        http_code: int,
        category: Optional[str] = None,
        description: Optional[str] = None,
        details: Optional[Union[str, Dict[str, str]]] = None,
    ):
        """Construct from individual properties."""
        self.http_code = http_code
        self.category = category
        self.description = description
        self.details = details

        super().__init__(http_code, category, description, details)

    @classmethod
    def from_response(cls, response: requests.Response) -> OnedataRESTError:
        """Construct from a requests response object."""
        http_code = response.status_code
        error_category = None
        error_description = None
        error_details = None

        try:
            error_json = response.json().get("error", {})
            error_category = error_json.get("id")
            error_description = error_json.get("description")
            error_details = error_json.get("details")
        except ValueError:
            # If response.json() raises a ValueError, it means the response is not JSON
            pass

        return cls(http_code, error_category, error_description, error_details)

    def __str__(self) -> str:
        """Describe error reason."""
        msg = f"HTTP {self.http_code}"

        if self.category:
            msg += f" [{self.category}]"

        if self.description:
            msg += f" {self.description}"

        if self.details:
            msg += f" ({self.details})"

        return msg



# coding: utf-8
"""Onezone REST API client."""

__author__ = "Bartosz Walkowicz"
__copyright__ = "Copyright (C) 2024 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"

import copy
import itertools
import sys
import time
import typing
from functools import lru_cache
from typing import Dict, Final, List, Optional, Tuple, Union, cast

from .errors import SpaceNotFoundError
from .http_client import HttpClient

if sys.version_info < (3, 11):
    from typing_extensions import TypeAlias, TypedDict
else:
    from typing import TypeAlias, TypedDict


ProviderId: TypeAlias = str

SpaceId: TypeAlias = str
SpaceName: TypeAlias = str
SpaceFQN: TypeAlias = str
"""
Fully qualified space name in the form: <SpaceName>@<SpaceId>
"""

SpaceSpecifier: TypeAlias = Union[SpaceName, SpaceFQN]


class SpaceSupportAttributes(TypedDict):
    """Relevant space support attributes.

    Refer to the API specification for more information:
    https://onedata.org/#/home/api/stable/onezone?anchor=operation/infer_access_token_scope
    """

    readonly: bool


class SpaceDetails(TypedDict):
    """Space details.

    Refer to the API specification for more information:
    https://onedata.org/#/home/api/stable/onezone?anchor=operation/infer_access_token_scope
    """

    name: str
    supports: Dict[ProviderId, SpaceSupportAttributes]


class ProviderDetails(TypedDict):
    """Provider details.

    Refer to the API specification for more information:
    https://onedata.org/#/home/api/stable/onezone?anchor=operation/infer_access_token_scope
    """

    name: str
    domain: str
    version: str
    online: bool


class DataAccessScope(TypedDict):
    """Data access scope info.

    Refer to the API specification for more information:
    https://onedata.org/#/home/api/stable/onezone?anchor=operation/infer_access_token_scope
    """

    readonly: bool
    spaces: Dict[SpaceId, SpaceDetails]
    providers: Dict[ProviderId, ProviderDetails]


class AccessTokenScope(TypedDict):
    """JSON object describing inferred data access scope and validity.

    Refer to the API specification for more information:
    https://onedata.org/#/home/api/stable/onezone?anchor=operation/infer_access_token_scope
    """

    validUntil: int
    dataAccessScope: DataAccessScope


_SPACE_ID_CACHE_SIZE_LIMIT: Final[int] = 512


class OnezoneRESTClient:
    """Custom REST client for Onezone REST basic operations API."""

    _host: str
    _token: str
    _http_client: HttpClient

    _token_scope_cache: Optional[AccessTokenScope] = None
    _token_scope_cache_valid_until_ns: int = 0
    _token_scope_cache_time_limit_ns: int = 2 * 10**9  # 2 seconds

    def __init__(self, host: str, token: str, *, verify_ssl: bool = True):
        """Construct OnezoneRESTClient instance."""
        self._host = host
        self._token = token
        self._http_client = HttpClient(verify_ssl=verify_ssl)

        # lru_cache cannot be used as decorator, as we want to have a separate
        # cache for each OnezoneRESTClient instance
        self._get_space_id_by_name_cache = lru_cache(
            maxsize=_SPACE_ID_CACHE_SIZE_LIMIT
        )(self._resolve_space_id_by_name)

    def __eq__(self, other: object) -> bool:
        """Compare 2 instances of OnezoneRESTClient."""
        if not isinstance(other, OnezoneRESTClient):
            return NotImplemented

        return (self._host == other._host) and (self._token == other._token)

    def __hash__(self) -> int:
        """Calculate a hash of a given instance of OnezoneRESTClient."""
        return hash(self._host) ^ hash(self._token)

    def build_url(self, path: str) -> str:
        """Build Onezone URL for specific path."""
        if not path.startswith("/"):
            path = "/" + path

        return f"https://{self._host}/api/v3/onezone{path}"

    def infer_token_scope(self) -> AccessTokenScope:
        """Get current token access scope."""
        self._ensure_cached_token_scope_is_up_to_date()
        return cast(AccessTokenScope, copy.deepcopy(self._token_scope_cache))

    def list_spaces(self) -> List[SpaceSpecifier]:
        """List all spaces available for the current token."""
        token_scope = self.infer_token_scope()
        space_registry = token_scope["dataAccessScope"]["spaces"]

        all_space_refs = [
            (space_details["name"], space_id)
            for space_id, space_details in space_registry.items()
        ]
        all_space_refs.sort()

        result: List[SpaceSpecifier] = []
        for space_name, refs_iter in itertools.groupby(
            all_space_refs, key=lambda x: x[0]
        ):
            refs_list = list(refs_iter)
            if len(refs_list) > 1:
                # Multiple spaces with the same name, use fully qualified name
                result.extend(pack_space_fqn(*space_ref) for space_ref in refs_list)
            else:
                # Only one space with this name, use the name alone
                result.append(space_name)

        return result

    def ensure_space_fqn(self, space_specifier: SpaceSpecifier) -> SpaceFQN:
        if is_fully_qualified_space_name(space_specifier):
            return space_specifier

        space_id = self._get_space_id_by_name(space_specifier)
        return pack_space_fqn(space_specifier, space_id)

    def get_space_id(self, space_specifier: SpaceSpecifier) -> SpaceId:
        """Get space id by specifier."""
        if is_fully_qualified_space_name(space_specifier):
            _, space_id = unpack_space_fqn(space_specifier)
            return space_id

        return self._get_space_id_by_name(space_specifier)

    def _get_space_id_by_name(self, space_name: SpaceName) -> SpaceId:
        # ensure space id cache is invalidated in case of token scope change
        self._ensure_cached_token_scope_is_up_to_date()
        return self._get_space_id_by_name_cache(space_name)

    def _resolve_space_id_by_name(self, space_name: SpaceName) -> SpaceId:
        access_token_scope = self.infer_token_scope()
        all_spaces = access_token_scope["dataAccessScope"]["spaces"]

        for space_id, space_details in all_spaces.items():
            if space_details["name"] == space_name:
                return space_id

        raise SpaceNotFoundError(space_name)

    def _ensure_cached_token_scope_is_up_to_date(self) -> None:
        if (
            self._token_scope_cache is None
            or time.time_ns() > self._token_scope_cache_valid_until_ns
        ):
            url = self.build_url("/tokens/infer_access_token_scope")
            result = self._http_client.post(url, {"token": self._token})
            access_token_scope = typing.cast(AccessTokenScope, result.json())

            if self._token_scope_cache != access_token_scope:
                # clear cache as it is possible that e.g. space name has changed
                self._get_space_id_by_name_cache.cache_clear()

            valid_until = time.time_ns() + self._token_scope_cache_time_limit_ns
            self._token_scope_cache = access_token_scope
            self._token_scope_cache_valid_until_ns = valid_until


def is_fully_qualified_space_name(space_specifier: SpaceSpecifier) -> bool:
    """Check if given space specifier if fully qualified."""
    return "@" in space_specifier


def pack_space_fqn(space_name: SpaceName, space_id: SpaceId) -> SpaceFQN:
    """Create space fully qualified name using space name and id."""
    return f"{space_name}@{space_id}"


def unpack_space_fqn(space_fqn: SpaceFQN) -> Tuple[SpaceName, SpaceId]:
    """Infer space name and id from fully qualified space name."""
    space_name, space_id = space_fqn.split("@")
    return space_name, space_id



# coding: utf-8
"""Definitions and utilities for handling file attributes.

Refer to the API specification for more information:
https://onedata.org/#/home/api/latest/oneprovider?anchor=operation/get_attrs
"""

__author__ = "Bartosz Walkowicz"
__copyright__ = "Copyright (C) 2024 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"

import sys
from typing import Any, Dict, Final, List, Literal, Optional, Tuple, Union, cast

from packaging.version import Version

from .errors import OnedataRESTError
from .provider_selector import SpaceSupportingProvider

if sys.version_info < (3, 11):
    from typing_extensions import TypeAlias, TypedDict
else:
    from typing import TypeAlias, TypedDict


_PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION: Final[Version] = Version("21.2.5")

_NOT_SUPPORTED_ATTR_ERROR_DETAILS_FMT: Final[str] = (
    "The provider chosen for this space ({domain}) is in version ({version}) "
    "that does not support the '{attr}' attribute (requires Oneprovider "
    "version >= {min_required_version})"
)

BasicFileAttrKey: TypeAlias = Literal[
    "fileId",
    "index",
    "type",
    "activePermissionsType",
    "posixPermissions",
    "acl",
    "name",
    "conflictingName",
    "path",
    "parentFileId",
    "displayGid",
    "displayUid",
    "atime",
    "mtime",
    "ctime",
    "size",
    "isFullyReplicatedLocally",
    "localReplicationRate",
    "originProviderId",
    "directShareIds",
    "ownerUserId",
    "hardlinkCount",
    "symlinkValue",
    "hasCustomMetadata",
    "effProtectionFlags",
    "effDatasetProtectionFlags",
    "effDatasetInheritancePath",
    "effQosInheritancePath",
    "aggregateQosStatus",
    "archiveRecallRootFileId",
]

XattrKey: TypeAlias = str
"""
File extended attribute in the form xattr.<NAME>
"""

FileAttrKey = Union[BasicFileAttrKey, XattrKey]

_DEPRECATED_BASIC_FILE_ATTR_KEYS: Dict[FileAttrKey, str] = {
    "fileId": "file_id",
    "index": "index",
    "type": "type",
    "posixPermissions": "mode",
    "name": "name",
    "parentFileId": "parent_id",
    "displayGid": "storage_group_id",
    "displayUid": "storage_user_id",
    "atime": "atime",
    "mtime": "mtime",
    "ctime": "ctime",
    "size": "size",
    "originProviderId": "provider_id",
    "directShareIds": "shares",
    "ownerUserId": "owner_id",
    "hardlinkCount": "hardlinks_count",
}

_DEFAULT_API_FILE_ATTR_KEYS: Final[List[FileAttrKey]] = [
    "fileId",
    "parentFileId",
    "name",
    "posixPermissions",
    "atime",
    "mtime",
    "ctime",
    "type",
    "size",
    "directShareIds",
    "index",
    "displayGid",
    "displayUid",
    "ownerUserId",
    "originProviderId",
    "hardlinkCount",
]

FileType: TypeAlias = Literal["REG", "DIR", "SYMLNK"]
ProtectionFlag: TypeAlias = Literal["data_protection", "metadata_protection"]
InheritancePath: TypeAlias = Literal[
    "none", "direct", "ancestor", "direct_and_ancestor"
]


class FileAttrsJson(TypedDict, total=False):
    """
    File attributes.

    Refer to the API specification for more information:
    https://onedata.org/#/home/api/latest/oneprovider?anchor=operation/get_attrs

    NOTE: dynamic fields like `xattr.*` are not typed due to limitations
    of typing machinery.
    """

    fileId: str
    index: str
    type: FileType
    activePermissionsType: Literal["posix", "acl"]
    posixPermissions: str
    acl: List[Any]
    name: str
    conflictingName: Optional[str]
    path: str
    parentFileId: Optional[str]
    displayGid: int
    displayUid: int
    atime: int
    mtime: int
    ctime: int
    size: Optional[int]
    isFullyReplicatedLocally: Optional[bool]
    localReplicationRate: float
    originProviderId: str
    directShareIds: List[str]
    ownerUserId: str
    hardlinkCount: int
    symlinkValue: Optional[str]
    hasCustomMetadata: bool
    effProtectionFlags: List[ProtectionFlag]
    effDatasetProtectionFlags: List[ProtectionFlag]
    effDatasetInheritancePath: InheritancePath
    effQosInheritancePath: InheritancePath
    aggregateQosStatus: Literal["fulfilled", "pending", "impossible"]
    archiveRecallRootFileId: Optional[str]


def build_http_get_file_attr_params(
    provider: SpaceSupportingProvider, requested_attr_keys: Optional[List[FileAttrKey]]
) -> Tuple[Optional[str], Optional[Dict[str, List[str]]]]:
    """Build query string and body for HTTP request to retrieve file attributes.

    NOTE: In case not all requested attributes are supported by selected
    provider exception will be raised.
    """
    qs = None
    body = None

    if requested_attr_keys is not None:
        if provider.version < _PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION:
            # Due to a bug it is not possible to fetch more than one specific
            # attribute for file using old API. Trying to do so results in empty
            # json returned. As a workaround all attributes are fetched (no qs)
            if len(requested_attr_keys) <= 1:
                qs = "&".join(
                    f"attribute={_get_deprecated_api_attr_key(provider, attr_key)}"
                    for attr_key in requested_attr_keys
                )
            else:
                # ensure all requested attributes are supported
                for attr_key in requested_attr_keys:
                    _get_deprecated_api_attr_key(provider, attr_key)
        else:
            body = {"attributes": requested_attr_keys}

    return qs, body


def build_http_list_children_attrs_params(
    provider: SpaceSupportingProvider, requested_attr_keys: Optional[List[FileAttrKey]]
) -> Tuple[Optional[str], Optional[Dict[str, List[str]]]]:
    """Build query string and body for HTTP request to retrieve file attributes.

    NOTE: In case not all requested attributes are supported by selected
    provider exception will be raised.
    """
    qs = None
    body = None

    if requested_attr_keys is not None:
        if provider.version < _PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION:
            qs = "&".join(
                f"attribute={_get_deprecated_api_attr_key(provider, attr_key)}"
                for attr_key in requested_attr_keys
            )
        else:
            body = {"attributes": requested_attr_keys}

    return qs, body


def _get_deprecated_api_attr_key(
    provider: SpaceSupportingProvider, attr_key: FileAttrKey
) -> str:
    deprecated_attr_key = _DEPRECATED_BASIC_FILE_ATTR_KEYS.get(attr_key)
    if deprecated_attr_key is not None:
        return deprecated_attr_key

    raise OnedataRESTError(
        http_code=400,
        category="posix",
        description="einval",
        details=_NOT_SUPPORTED_ATTR_ERROR_DETAILS_FMT.format(
            domain=provider.domain,
            version=provider.version,
            attr=attr_key,
            min_required_version=_PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION,
        ),
    )


def normalize_file_attrs_json(
    provider: SpaceSupportingProvider,
    requested_attr_keys: Optional[List[FileAttrKey]],
    file_attrs_json: Dict[str, Any],
) -> FileAttrsJson:
    """Translate attributes api keys to current ones for older providers."""
    if not requested_attr_keys:
        # default api attrs
        requested_attr_keys = _DEFAULT_API_FILE_ATTR_KEYS

    if provider.version < _PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION:
        sanitized_file_attrs_json = {
            attr_key: file_attrs_json[_DEPRECATED_BASIC_FILE_ATTR_KEYS[attr_key]]
            for attr_key in requested_attr_keys
        }
    else:
        sanitized_file_attrs_json = {
            attr_key: file_attrs_json[attr_key] for attr_key in requested_attr_keys
        }

    return cast(FileAttrsJson, sanitized_file_attrs_json)



# coding: utf-8
"""Provider selector utilities."""

__author__ = "Bartosz Walkowicz"
__copyright__ = "Copyright (C) 2024 ACK CYFRONET AGH"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"

import logging
import sys
import time
from datetime import datetime
from typing import Dict, Final, Iterator, List, NamedTuple, Optional, Union

from packaging.version import Version, parse

from .onezone_rest_client import (
    OnezoneRESTClient,
    ProviderDetails,
    ProviderId,
    SpaceSpecifier,
    SpaceSupportAttributes,
)

if sys.version_info < (3, 11):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias


_logger = logging.getLogger(__name__)
_logger.addHandler(logging.NullHandler())

_MIN_SUPPORTED_PROVIDER_VERSION: Final[Version] = Version("21.2.1")

ProviderDomain: TypeAlias = str
ProviderSpecifier: TypeAlias = Union[ProviderId, ProviderDomain]


class SpaceSupportingProvider(NamedTuple):
    """Provider relevant attributes."""

    id: str
    version: Version
    domain: str
    online: bool
    readonly_support: bool


class ProviderSelector:
    """Selector responsible for choosing available provider(s) for space."""

    preferred_providers: List[str]

    _cache_size_limit: int = 512
    _provider_for_space_cache: Dict[SpaceSpecifier, SpaceSupportingProvider]
    _provider_blacklist_cache: Dict[ProviderId, int]
    _blacklist_time_limit_ns: int = 30 * 10**9  # 30 seconds

    def __init__(
        self, *, preferred_providers: Optional[List[ProviderSpecifier]] = None
    ) -> None:
        """Construct ProviderSelector instance."""
        self.preferred_providers = preferred_providers or []
        self._provider_blacklist_cache = {}
        self._provider_for_space_cache = {}

    def is_blacklisted(self, provider_id: ProviderId) -> bool:
        """Blacklist specified provider for a short while."""
        if provider_id not in self._provider_blacklist_cache:
            return False

        blacklist_time_end = self._provider_blacklist_cache[provider_id]
        if blacklist_time_end > time.time_ns():
            return True

        del self._provider_blacklist_cache[provider_id]
        return False

    def blacklist(self, provider: SpaceSupportingProvider) -> None:
        """Check if specified provider is blacklisted."""
        blacklist_time_end_ns = time.time_ns() + self._blacklist_time_limit_ns

        _logger.debug(
            "Blacklisting provider '%s' (id: %s) until %s",
            provider.domain,
            provider.id,
            datetime.fromtimestamp(blacklist_time_end_ns // 1e9),
        )

        if len(self._provider_blacklist_cache) > self._cache_size_limit:
            self._provider_blacklist_cache = {provider.id: blacklist_time_end_ns}
        else:
            self._provider_blacklist_cache[provider.id] = blacklist_time_end_ns

    def iter_available_space_providers(
        self,
        space_specifier: SpaceSpecifier,
        *,
        oz_rest_client: OnezoneRESTClient,
        except_readonly: bool = False,
    ) -> Iterator[SpaceSupportingProvider]:
        """Iterate over online and not blacklisted space providers."""
        space_fqn = oz_rest_client.ensure_space_fqn(space_specifier)

        cache_key = space_fqn
        yield from self._fetch_provider_from_cache(cache_key, except_readonly)

        if except_readonly:
            cache_key = f"{space_fqn}#not_readonly"
            yield from self._fetch_provider_from_cache(cache_key, except_readonly)

        if len(self._provider_for_space_cache) >= self._cache_size_limit:
            # clear cache
            self._provider_for_space_cache = {}

        for provider in self.list_available_space_providers(
            space_fqn,
            oz_rest_client=oz_rest_client,
            except_readonly=except_readonly,
        ):
            _logger.debug(
                "Designating provider '%s' (id: %s) to handle requests for space '%s'",
                provider.domain,
                provider.id,
                space_fqn,
            )
            self._provider_for_space_cache[cache_key] = provider
            yield provider

    def list_available_space_providers(
        self,
        space_specifier: SpaceSpecifier,
        *,
        oz_rest_client: OnezoneRESTClient,
        except_readonly: bool = False,
    ) -> List[SpaceSupportingProvider]:
        """List online and not not blacklisted space providers."""
        space_id = oz_rest_client.get_space_id(space_specifier)

        access_token_scope = oz_rest_client.infer_token_scope()
        all_providers = access_token_scope["dataAccessScope"]["providers"]
        space_details = access_token_scope["dataAccessScope"]["spaces"][space_id]

        preferred_supporting_providers = []
        remaining_supporting_providers = []

        for provider_id, support_attributes in space_details["supports"].items():
            provider = self._build_space_supporting_provider(
                provider_id, support_attributes, all_providers[provider_id]
            )
            if not self._is_provider_available(provider, except_readonly):
                continue

            try:
                index = next(
                    i
                    for i, provider_specifier in enumerate(self.preferred_providers)
                    if provider_specifier in (provider.id, provider.domain)
                )
            except StopIteration:
                remaining_supporting_providers.append(provider)
            else:
                preferred_supporting_providers.append((index, provider))

        preferred_supporting_providers.sort()
        supporting_providers = [
            provider for _, provider in preferred_supporting_providers
        ]

        remaining_supporting_providers.sort(key=lambda x: x.version, reverse=True)
        supporting_providers.extend(remaining_supporting_providers)

        return supporting_providers

    @staticmethod
    def _build_space_supporting_provider(
        provider_id: ProviderId,
        support_attributes: SpaceSupportAttributes,
        provider_details: ProviderDetails,
    ) -> SpaceSupportingProvider:
        provider = SpaceSupportingProvider(
            id=provider_id,
            version=parse(provider_details["version"]),
            domain=provider_details["domain"],
            online=provider_details["online"],
            readonly_support=support_attributes["readonly"],
        )
        return provider

    def _is_provider_available(
        self, provider: SpaceSupportingProvider, except_readonly: bool
    ) -> bool:
        if self.is_blacklisted(provider.id):
            return False

        if except_readonly and provider.readonly_support:
            return False

        if not provider.online:
            return False

        if provider.version < _MIN_SUPPORTED_PROVIDER_VERSION:
            return False

        return True

    def _fetch_provider_from_cache(
        self, cache_key: str, except_readonly: bool
    ) -> Iterator[SpaceSupportingProvider]:
        provider = self._provider_for_space_cache.get(cache_key)
        if provider is not None:
            if (not except_readonly) or (
                except_readonly and not provider.readonly_support
            ):
                if not self.is_blacklisted(provider.id):
                    yield provider

                del self._provider_for_space_cache[cache_key]



# coding: utf-8
"""Test OnedataFileRESTClient methods."""

import json
import os
import random
import time
from contextlib import contextmanager
from typing import get_args

import requests
from packaging.version import Version  # type: ignore

import pytest
from onedatafilerestclient import OnedataFileRESTClient, OnedataRESTError
from onedatafilerestclient.errors import (
    NoAvailableProviderForSpaceError,
    ReadonlyTokenError,
    SpaceNotFoundError,
)
from onedatafilerestclient.file_attributes import (
    _DEPRECATED_BASIC_FILE_ATTR_KEYS,
    BasicFileAttrKey,
)

from . import (
    PROVIDER_KRK_DOMAIN,
    PROVIDER_PAR_DOMAIN,
    SPACE_KRK_PAR_NAME,
    SPACE_NO_SUPPORT_NAME,
    SPACE_PAR_NAME,
    ZONE_DOMAIN,
)
from .utils import random_bytes, random_int, random_path, random_str


@pytest.fixture(name="client_verifying_ssl")
def fixture_client_verifying_ssl(onezone_ip, onezone_admin_token):
    """Create OnedataFileRESTClient instance."""
    return OnedataFileRESTClient(onezone_ip, onezone_admin_token, verify_ssl=True)


@pytest.fixture(name="client")
def fixture_client(onezone_ip, onezone_admin_token):
    """Create OnedataFileRESTClient instance."""
    return OnedataFileRESTClient(onezone_ip, onezone_admin_token, verify_ssl=False)


@pytest.fixture(name="client_ro")
def fixture_client_ro(onezone_ip, onezone_readonly_token):
    """Create readonly OnedataFileRESTClient instance."""
    return OnedataFileRESTClient(onezone_ip, onezone_readonly_token, verify_ssl=False)


@pytest.fixture(name="client_krakow")
def fixture_client_krakow(onezone_ip, onezone_admin_token):
    """Create OnedataFileRESTClient instance bound to 'krakow' provider."""
    return OnedataFileRESTClient(
        onezone_ip, onezone_admin_token, [PROVIDER_KRK_DOMAIN], verify_ssl=False
    )


@pytest.fixture(name="client_ro_krakow")
def fixture_client_ro_krakow(onezone_ip, onezone_readonly_token):
    """Create readonly OnedataFileRESTClient instance bound to 'krakow'."""
    return OnedataFileRESTClient(
        onezone_ip, onezone_readonly_token, [PROVIDER_KRK_DOMAIN], verify_ssl=False
    )


@pytest.fixture(name="client_paris")
def fixture_client_paris(onezone_ip, onezone_admin_token):
    """Create OnedataFileRESTClient instance bound to 'krakow' provider."""
    return OnedataFileRESTClient(
        onezone_ip, onezone_admin_token, [PROVIDER_PAR_DOMAIN], verify_ssl=False
    )


def test_ssl_verification(client_verifying_ssl: OnedataFileRESTClient):
    """Test 'OnedataFileRESTClient' respects 'verify_ssl' flag."""
    with pytest.raises(requests.exceptions.SSLError):
        assert client_verifying_ssl.list_spaces()


def test_list_spaces(client: OnedataFileRESTClient):
    """Test 'list_spaces' method.

    Test listing user spaces and proper handling of duplicate space names
    (the space with no support is duplicated, see the test_env_config.yaml)
    """
    token_scope = client.get_token_scope()
    space_registry = token_scope["dataAccessScope"]["spaces"]

    exp_space_list = []
    for space_id, space_details in space_registry.items():
        space_specifier = space_details["name"]
        if space_specifier == SPACE_NO_SUPPORT_NAME:
            space_specifier = _pack_space_fqn(space_specifier, space_id)

        exp_space_list.append(space_specifier)

    exp_space_list.sort()

    assert exp_space_list == sorted(client.list_spaces())


def test_get_space_id(onezone_ip, onezone_admin_token):
    """Test 'get_space_id' method."""
    client = OnedataFileRESTClient(
        onezone_ip, onezone_admin_token, [PROVIDER_KRK_DOMAIN], verify_ssl=False
    )
    client._oz_client._space_id_cache_size_limit = 2  # pylint: disable=W0212

    space_par_id = client.get_space_id(SPACE_PAR_NAME)

    # space_fqn resolution should not be cached
    space_par_fqn = _pack_space_fqn(SPACE_PAR_NAME, space_par_id)
    assert client.get_space_id(space_par_fqn) == space_par_id

    space_par_new_name = random_str()
    _rename_space(onezone_admin_token, space_par_id, space_par_new_name)

    try:
        # changes should not be reflected automatically since token is cached
        time.sleep(1)
        assert space_par_id == client.get_space_id(SPACE_PAR_NAME)
        with pytest.raises(SpaceNotFoundError):
            client.get_space_id(space_par_new_name)

        # but after cache expires the changes should be reflected
        time.sleep(3)
        with pytest.raises(SpaceNotFoundError):
            client.get_space_id(SPACE_PAR_NAME)

        assert space_par_id == client.get_space_id(space_par_new_name)

        # using fqn returns id from fqn so no resolution is made and old name
        # still works
        assert client.get_space_id(space_par_fqn) == space_par_id
    finally:
        _rename_space(onezone_admin_token, space_par_id, SPACE_PAR_NAME)
        time.sleep(3)


def test_provider_selector(
    onezone_ip, onezone_admin_token, client_krakow, client_paris
):
    """Test provider fallback on connection error."""
    providers = [PROVIDER_KRK_DOMAIN, PROVIDER_PAR_DOMAIN]
    random.shuffle(providers)
    first_choice_provider, second_choice_provider = providers

    client = OnedataFileRESTClient(
        onezone_ip,
        onezone_admin_token,
        [_random_provider_specifier(first_choice_provider)],
        verify_ssl=False,
    )

    # pylint: disable=W0212
    client._provider_selector._blacklist_time_limit_ns = 1 * 10**9

    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)
    file_id = _create_and_sync_file_in_space(
        space_specifier, client_krakow, client_paris
    )

    def assert_selected_provider(provider_domain):
        client.get_attributes(space_specifier, file_id=file_id)
        assert _get_selected_provider(client, space_specifier).domain == provider_domain

        random_mode = random.choice(["777", "775", "773", "771", "770"])
        client.set_attributes(space_specifier, {"mode": random_mode}, file_id=file_id)
        assert _get_selected_provider(client, space_specifier).domain == provider_domain

    # provider 'first_choice_provider' is chosen with accordance to
    # preferred providers
    assert_selected_provider(first_choice_provider)

    # with connection error raised 'first_choice_provider' should be
    # blacklisted for a while and next in line provider
    # - second_choice_provider - should be selected
    with _mock_http_client([first_choice_provider]):
        assert_selected_provider(second_choice_provider)

    assert_selected_provider(second_choice_provider)

    # with connection error raised by 'second_choice_provider' there should
    # be no available providers left
    with _mock_http_client([second_choice_provider]):
        with pytest.raises(NoAvailableProviderForSpaceError) as exc_info:
            client.get_attributes(space_specifier, file_id=file_id)

        assert exc_info.value.args == (space_specifier,)

    # even without mock providers should still be blacklisted
    with pytest.raises(NoAvailableProviderForSpaceError) as exc_info:
        client.set_attributes(space_specifier, {"mode": "777"}, file_id=file_id)

    assert exc_info.value.args == (space_specifier,)

    # but after blacklist time ends 'first_choice_provider' should be
    # again selected
    time.sleep(2)
    assert_selected_provider(first_choice_provider)


def test_provider_selector_with_readonly_provider(
    onezone_ip, onezone_admin_token, client_krakow, client_paris
):
    """Test readonly provider fallback on connection error."""
    providers = [PROVIDER_KRK_DOMAIN, PROVIDER_PAR_DOMAIN]
    random.shuffle(providers)
    first_choice_provider, second_choice_provider = providers
    first_choice_provider_id = _get_provider_id(first_choice_provider)

    client = OnedataFileRESTClient(
        onezone_ip,
        onezone_admin_token,
        [_random_provider_specifier(first_choice_provider)],
        verify_ssl=False,
    )

    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)
    space_id = client.get_space_id(space_specifier)
    file_id = _create_and_sync_file_in_space(
        space_specifier, client_krakow, client_paris
    )

    # pylint: disable=W0212
    client._provider_selector._blacklist_time_limit_ns = 1 * 10**9  # 1 second

    # mock token scope so that 'first_choice_provider' has readonly support
    _patch_provider_readonly_support(client, space_id, first_choice_provider_id)

    def get_selected_provider_domain(except_readonly):
        return _get_selected_provider(
            client, space_specifier, except_readonly=except_readonly
        ).domain

    def assert_selected_provider(read_provider_domain, write_provider_domain=None):
        client.get_attributes(space_specifier, file_id=file_id)
        assert get_selected_provider_domain(False) == read_provider_domain

        if write_provider_domain is None:
            write_provider_domain = read_provider_domain
        random_mode = random.choice(["777", "775", "773", "771", "770"])
        client.set_attributes(space_specifier, {"mode": random_mode}, file_id=file_id)
        assert get_selected_provider_domain(True) == write_provider_domain

    # provider 'first_choice_provider' is chosen with accordance to
    # preferred providers but only for read (readonly support)
    assert_selected_provider(first_choice_provider, second_choice_provider)

    # with connection error raised 'first_choice_provider' should be
    # blacklisted for a while and next in line provider
    # - second_choice_provider - should be selected
    with _mock_http_client([first_choice_provider]):
        assert_selected_provider(second_choice_provider)

    assert_selected_provider(second_choice_provider)

    # with connection error raised by 'second_choice_provider' there should
    # be no available providers left
    with _mock_http_client([second_choice_provider]):
        with pytest.raises(NoAvailableProviderForSpaceError) as exc_info:
            client.get_attributes(space_specifier, file_id=file_id)

        assert exc_info.value.args == (space_specifier,)

    # even without mock providers should still be blacklisted
    with pytest.raises(NoAvailableProviderForSpaceError) as exc_info:
        client.set_attributes(space_specifier, {"mode": "777"}, file_id=file_id)

    assert exc_info.value.args == (space_specifier,)

    # but after blacklist time ends 'first_choice_provider' should be
    # again selected for read and 'second_choice_provider' for write
    time.sleep(2)
    assert_selected_provider(first_choice_provider, second_choice_provider)

    # with connection error raised by 'second_choice_provider' there should
    # be no available providers for write but read should still work
    with _mock_http_client([second_choice_provider]):
        with pytest.raises(NoAvailableProviderForSpaceError) as exc_info:
            client.set_attributes(space_specifier, {"mode": "777"}, file_id=file_id)

        assert exc_info.value.args == (space_specifier,)

    client.get_attributes(space_specifier, file_id=file_id)
    assert get_selected_provider_domain(False) == first_choice_provider


def test_provider_selector_with_offline_provider(onezone_ip, onezone_admin_token):
    """Test offline provider should be omitted."""
    providers = [PROVIDER_KRK_DOMAIN, PROVIDER_PAR_DOMAIN]
    random.shuffle(providers)
    first_choice_provider, second_choice_provider = providers
    first_choice_provider_id = _get_provider_id(first_choice_provider)

    client = OnedataFileRESTClient(
        onezone_ip,
        onezone_admin_token,
        [_random_provider_specifier(first_choice_provider)],
        verify_ssl=False,
    )
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    _patch_provider_offline(client, first_choice_provider_id)

    client.get_attributes(space_specifier)
    assert (
        _get_selected_provider(client, space_specifier).domain == second_choice_provider
    )


def test_provider_selector_with_dummy_provider(onezone_ip, onezone_admin_token):
    """Test dummy provider should be omitted."""
    client_1 = OnedataFileRESTClient(
        onezone_ip,
        onezone_admin_token,
        ["dummy.org", _random_provider_specifier(PROVIDER_KRK_DOMAIN)],
        verify_ssl=False,
    )
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client_1)

    client_1.get_attributes(space_specifier)
    assert (
        _get_selected_provider(client_1, space_specifier).domain == PROVIDER_KRK_DOMAIN
    )

    client_2 = OnedataFileRESTClient(
        onezone_ip,
        onezone_admin_token,
        ["dummyId", _random_provider_specifier(PROVIDER_PAR_DOMAIN)],
        verify_ssl=False,
    )
    client_2.get_attributes(space_specifier)
    assert (
        _get_selected_provider(client_2, space_specifier).domain == PROVIDER_PAR_DOMAIN
    )


def test_get_file_id(client: OnedataFileRESTClient):
    """Test 'get_file_id' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    file_path = random_path()
    file_id = client.create_file(space_specifier, file_path, create_parents=True)

    assert file_id == client.get_file_id(space_specifier, file_path=file_path)


def test_get_attributes_for_space_dir(client: OnedataFileRESTClient):
    """Test 'get_attributes' method for space directory."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)
    space_attrs = client.get_attributes(space_specifier)

    assert SPACE_KRK_PAR_NAME == space_attrs["name"]


def test_get_attributes_for_file(client: OnedataFileRESTClient):
    """Test 'get_attributes' method for regular file."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    file_path = random_path()
    file_name = file_path.rsplit("/", maxsplit=1)[-1]
    file_id = client.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    file_attrs = client.get_attributes(space_specifier, **file_selector)

    assert file_name == file_attrs["name"]
    assert file_id == file_attrs["fileId"]


def test_get_selected_attributes(client: OnedataFileRESTClient):
    """Test 'get_attributes' method with random attributes."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    # Set min version to lower that testing providers to allow all new attrs
    with _mock_min_provider_version_supporting_new_file_attrs(Version("20.1.1")):

        requested_attr_keys = random.sample(get_args(BasicFileAttrKey), 5)
        space_attrs = client.get_attributes(
            space_specifier, attributes=requested_attr_keys
        )
        assert sorted(requested_attr_keys) == sorted(space_attrs.keys())

    # Set min version to higher that testing providers to disallow new attrs
    with _mock_min_provider_version_supporting_new_file_attrs(Version("300.1.1")):
        requested_attr_keys = random.sample(_DEPRECATED_BASIC_FILE_ATTR_KEYS.keys(), 5)
        requested_attr_keys.append("acl")

        with pytest.raises(OnedataRESTError) as exc_info:
            client.get_attributes(space_specifier, attributes=requested_attr_keys)

        selected_provider = _get_selected_provider(client, space_specifier)
        error = exc_info.value
        assert error.http_code == 400
        assert error.category == "posix"
        assert error.details == (
            f"The provider chosen for this space ({selected_provider.domain}) is in version "
            f"({selected_provider.version}) that does not support the 'acl' "
            "attribute (requires Oneprovider version >= 300.1.1)"
        )
        assert error.description == "einval"

        space_attrs = client.get_attributes(
            space_specifier, attributes=requested_attr_keys[:-1]
        )
        exp_attrs = sorted(requested_attr_keys[:-1])
        assert exp_attrs == sorted(space_attrs.keys())


def test_set_attributes(client: OnedataFileRESTClient):
    """Test 'set_attributes' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    file_path = random_path()
    file_id = client.create_file(
        space_specifier, file_path, mode=0o775, create_parents=True
    )
    file_selector = _random_file_selector(file_id, file_path)

    file_attrs = client.get_attributes(space_specifier, **file_selector)
    assert file_attrs["posixPermissions"] == "775"

    client.set_attributes(space_specifier, {"mode": "553"}, **file_selector)
    file_attrs = client.get_attributes(space_specifier, **file_selector)

    assert file_attrs["posixPermissions"] == "553"


def test_list_children(client: OnedataFileRESTClient):
    """Test 'list_children' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    file_count = random_int(20, 50)
    dir_path = random_path()

    exp_children = []
    for _ in range(file_count):
        file_name = random_str(random_int(lower_bound=10))
        file_path = os.path.join(dir_path, file_name)
        client.create_file(space_specifier, file_path, create_parents=True)

        exp_children.append(
            {
                "name": file_name,
                "type": "REG",
            }
        )

    token = None
    limit = random_int(4, 10)
    dir_id = client.get_file_id(space_specifier, file_path=dir_path)
    dir_selector = _random_file_selector(dir_id, dir_path)

    children = []
    while True:
        result = client.list_children(
            space_specifier, limit=limit, continuation_token=token, **dir_selector
        )
        children.extend(result["children"])
        if result["isLast"]:
            break

        token = result["nextPageToken"]

    def sort_files(files):
        return sorted(files, key=lambda d: d["name"])

    assert sort_files(children) == sort_files(exp_children)


def test_get_file_content(client_krakow: OnedataFileRESTClient):
    """Test 'get_file_content' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client_krakow)

    file_path = random_path()
    file_id = client_krakow.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    file_size = 1024
    exp_file_content = random_bytes(file_size)
    client_krakow.put_file_content(
        space_specifier, data=exp_file_content, **file_selector
    )

    assert exp_file_content == client_krakow.get_file_content(
        space_specifier, **file_selector
    )

    assert exp_file_content[100:300] == client_krakow.get_file_content(
        space_specifier, offset=100, size=200, **file_selector
    )


def test_get_file_content_with_readonly_token_on_the_same_provider(
    client_krakow: OnedataFileRESTClient, client_ro_krakow: OnedataFileRESTClient
):
    """Test 'get_file_content' method using readonly token."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client_krakow)

    file_path = random_path()
    file_id = client_krakow.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    file_content = random_bytes(1024)
    client_krakow.put_file_content(space_specifier, data=file_content, **file_selector)

    assert (
        client_ro_krakow.get_file_content(space_specifier, **file_selector)
        == file_content
    )


def test_enoent_space(client: OnedataFileRESTClient):
    """Test 'get_file_content' on non-existing space."""
    with pytest.raises(SpaceNotFoundError) as exc_info:
        client.get_file_content("NO_SUCH_SPACE", file_path=random_path())

    assert exc_info.value.args == ("NO_SUCH_SPACE",)


def test_enoent_file(client: OnedataFileRESTClient):
    """Test 'get_file_content' on non-existing file."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)
    file_path = random_path()

    with pytest.raises(OnedataRESTError) as exc_info:
        client.get_file_content(space_specifier, file_path=file_path)

    e = exc_info.value
    assert e.http_code == 400
    assert e.category == "posix"
    assert e.description == "Operation failed with POSIX error: enoent."
    assert e.details == {"errno": "enoent"}


def test_eacces_file(onezone_ip, onezone_space_member_token):
    """Test 'get_file_content' on non-existing file."""
    client = OnedataFileRESTClient(
        onezone_ip, onezone_space_member_token, verify_ssl=False
    )

    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    file_path = random_path()
    file_id = client.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    client.set_attributes(space_specifier, {"mode": "000"}, **file_selector)

    print(_get_selected_provider(client, space_specifier).domain)
    with pytest.raises(OnedataRESTError) as exc_info:
        file_content = random_bytes(1024)
        client.put_file_content(space_specifier, data=file_content, **file_selector)
        print(_get_selected_provider(client, space_specifier).domain)

    e = exc_info.value
    assert e.http_code == 400
    assert e.category == "posix"
    assert e.description == "Operation failed with POSIX error: eacces."
    assert e.details == {"errno": "eacces"}


def test_iter_file_content(client_krakow: OnedataFileRESTClient):
    """Test 'iter_file_content' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client_krakow)

    file_path = random_path()
    file_id = client_krakow.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    file_content = random_bytes(1024)
    client_krakow.put_file_content(space_specifier, data=file_content, **file_selector)

    chunk_size = random_int(4, 100)
    stream = client_krakow.iter_file_content(
        space_specifier, chunk_size=chunk_size, **file_selector
    )

    buff = b""
    for chunk in stream:
        assert len(chunk) <= chunk_size
        buff += chunk

    assert buff == file_content


def test_put_file_content(client: OnedataFileRESTClient):
    """Test 'put_file_content' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    file_path = random_path()
    file_id = client.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    file_size = 100
    rand_bytes = random_bytes(file_size)
    client.put_file_content(
        space_specifier, data=rand_bytes, offset=100, **file_selector
    )
    exp_file_content = 100 * b"\0" + rand_bytes

    assert exp_file_content == client.get_file_content(space_specifier, **file_selector)


def test_create_file(client: OnedataFileRESTClient):
    """Test 'create_file' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    file_path = random_path()
    file_id = client.create_file(
        space_specifier, file_path, file_type="REG", create_parents=True
    )
    file_selector = _random_file_selector(file_id, file_path)

    file_content = random_bytes(1024)
    client.put_file_content(space_specifier, data=file_content, **file_selector)

    content = client.get_file_content(space_specifier, **file_selector)

    assert content == file_content


def test_remove(client: OnedataFileRESTClient):
    """Test 'remove' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    dir_path = random_path()
    file_path = os.path.join(dir_path, random_str())
    file_id = client.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    res = client.list_children(space_specifier, file_path=dir_path)
    assert len(res["children"]) == 1

    client.remove(space_specifier, **file_selector)

    res = client.list_children(space_specifier, file_path=dir_path)
    assert len(res["children"]) == 0


def test_remove_with_readonly_token(
    client_krakow: OnedataFileRESTClient, client_ro_krakow: OnedataFileRESTClient
):
    """Test 'remove' method using readonly token."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client_krakow)

    dir_path = random_path()
    file_path = os.path.join(dir_path, random_str())
    file_id = client_krakow.create_file(space_specifier, file_path, create_parents=True)
    file_selector = _random_file_selector(file_id, file_path)

    file_content = random_bytes(1024)
    client_krakow.put_file_content(space_specifier, data=file_content, **file_selector)

    with pytest.raises(ReadonlyTokenError):
        client_ro_krakow.remove(space_specifier, **file_selector)


def test_move(client: OnedataFileRESTClient):
    """Test 'move' method."""
    space_specifier = _random_space_specifier(SPACE_KRK_PAR_NAME, client)

    dir_path = random_path()
    file_path = os.path.join(dir_path, random_str())
    client.create_file(space_specifier, file_path, create_parents=True)

    target_test_dir = random_path()
    client.create_file(
        space_specifier, target_test_dir, file_type="DIR", create_parents=True
    )

    target_file_path = os.path.join(target_test_dir, random_str())

    res = client.list_children(space_specifier, file_path=dir_path)
    assert len(res["children"]) == 1

    client.move(space_specifier, file_path, space_specifier, target_file_path)

    res = client.list_children(space_specifier, file_path=dir_path)
    assert len(res["children"]) == 0

    res = client.list_children(space_specifier, file_path=target_test_dir)
    assert len(res["children"]) == 1


def _random_space_specifier(space_name, client):
    space_fqn = _get_space_fqn(space_name, client)
    return random.choice([space_name, space_fqn])


def _get_space_fqn(space_name, client):
    return _pack_space_fqn(space_name, client.get_space_id(space_name))


def _pack_space_fqn(space_name, space_id):
    return f"{space_name}@{space_id}"


def _random_file_selector(file_id, file_path):
    if random.choice([True, False]):
        return {"file_id": file_id}

    return {"file_path": file_path}


def _random_provider_specifier(domain):
    # return domain if random.choice([True, False]) else _get_provider_id(domain)
    return _get_provider_id(domain)


def _get_selected_provider(client, space_specifier, *, except_readonly=False):
    # pylint: disable=W0212
    oz_client = client._oz_client
    provider = next(
        client._provider_selector.iter_available_space_providers(
            space_specifier, oz_rest_client=oz_client, except_readonly=except_readonly
        )
    )
    return provider


def _create_and_sync_file_in_space(space_specifier, client_krakow, client_paris):
    file_id = client_krakow.create_file(
        space_specifier, random_str(), create_parents=True
    )
    for _ in range(10):
        result = client_paris.list_children(space_specifier, attributes=["fileId"])
        if {"fileId": file_id} in result["children"]:
            break
        time.sleep(1)
    else:
        assert False, "Failed to create and sync file between providers"

    return file_id


def _rename_space(token, space_id, new_name):
    result = requests.patch(
        f"https://{ZONE_DOMAIN}/api/v3/onezone/spaces/{space_id}",
        headers={"X-Auth-Token": token, "Content-type": "application/json"},
        data=json.dumps({"name": new_name}),
        verify=False,
    )
    if not result.ok:
        raise OnedataRESTError.from_response(result)


def _get_provider_id(host: str) -> str:
    result = requests.get(
        f"https://{host}/api/v3/oneprovider/configuration",
        verify=False,
    )
    return result.json()["providerId"]


def _patch_provider_readonly_support(client, space_id, provider_id):
    # pylint: disable=W0212
    client._oz_client._token_scope_cache_time_limit_ns = 30 * 10**9  # 30 seconds
    access_token_scope = client.get_token_scope()
    space_details = access_token_scope["dataAccessScope"]["spaces"][space_id]
    space_details["supports"][provider_id]["readonly"] = True
    client._oz_client._token_scope_cache = access_token_scope


def _patch_provider_offline(client, provider_id):
    # pylint: disable=W0212
    client._oz_client._token_scope_cache_time_limit_ns = 30 * 10**9  # 30 seconds
    access_token_scope = client.get_token_scope()
    access_token_scope["dataAccessScope"]["providers"][provider_id]["online"] = False
    client._oz_client._token_scope_cache = access_token_scope


@contextmanager
def _mock_http_client(raise_connection_error_for_provider_domains):
    # pylint: disable=C0415
    from onedatafilerestclient.http_client import HttpClient

    # pylint: disable=W0212
    original_send_request_method = HttpClient._send_request

    def mock_send_request(self, method, url, *args, **kwargs):
        for domain in raise_connection_error_for_provider_domains:
            if domain in url:
                raise requests.exceptions.ConnectionError()

        return original_send_request_method(self, method, url, *args, **kwargs)

    try:
        # pylint: disable=W0212
        HttpClient._send_request = mock_send_request
        yield
    finally:
        # pylint: disable=W0212
        HttpClient._send_request = original_send_request_method


@contextmanager
def _mock_min_provider_version_supporting_new_file_attrs(mock_version):
    # pylint: disable=C0415
    import onedatafilerestclient.file_attributes as fa

    # pylint: disable=W0212
    original_version = fa._PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION
    try:
        # pylint: disable=W0212
        fa._PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION = mock_version
        yield
    finally:
        # pylint: disable=W0212
        fa._PROVIDER_SUPPORTING_CURRENT_API_KEY_MIN_VERSION = original_version



# coding: utf-8
"""onedatafilerestclient module."""

__author__ = "Bartek Kryza"
__copyright__ = "Copyright (C) 2023 Onedata"
__license__ = "This software is released under the MIT license cited in LICENSE.txt"

__all__ = ["OnedataRESTError", "OnedataFileRESTClient"]

from .errors import OnedataRESTError  # noqa
from .onedata_file_rest_client import OnedataFileRESTClient  # noqa



#!/usr/bin/env python
"""Onedata REST file API client."""

from setuptools import setup

__version__ = "21.2.5.1"

CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
    "Programming Language :: Python :: 3.12",
    "Topic :: System :: Filesystems",
]

with open("README.md", "rt") as f:
    DESCRIPTION = f.read()

REQUIREMENTS = ["packaging", "requests", "typing-extensions>=4.3.0"]

setup(
    name="onedatafilerestclient",
    version=__version__,
    description="Onedata REST file API client",
    long_description=DESCRIPTION,
    long_description_content_type="text/markdown",
    author="Bartek Kryza",
    author_email="bkryza@gmail.com",
    license="MIT",
    classifiers=CLASSIFIERS,
    python_requires=">=3.8",
    install_requires=REQUIREMENTS,
    packages=["onedatafilerestclient"],
    package_data={"onedatafilerestclient": ["py.typed"]},
    include_package_data=True,
    keywords=["Onedata"],
    test_suite="nose.collector",
    url="https://github.com/onedata/onedatafilerestclient",
)
